@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2001-2004 Sistina Software, Inc. All rights reserved.
- * Copyright (C) 2004-2005 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2004-2007 Red Hat, Inc. All rights reserved.
  *
  * This file is part of the device-mapper userspace tools.
  *
@@ -16,7 +16,6 @@
 #include "lib.h"
 #include "libdm-targets.h"
 #include "libdm-common.h"
-#include "libdm-file.h"
 
 #ifdef DM_COMPAT
 #  include "libdm-compat.h"
@@ -140,7 +139,7 @@ static int _get_proc_number(const char *file, const char *name,
 	uint32_t num;
 
 	if (!(fl = fopen(file, "r"))) {
-		log_error("%s: fopen failed: %s", file, strerror(errno));
+		log_sys_error("fopen", file);
 		return 0;
 	}
 
@@ -150,7 +149,7 @@ static int _get_proc_number(const char *file, const char *name,
 				if (number) {
 					*number = num;
 					if (fclose(fl))
-						log_error("%s: fclose failed: %s", file, strerror(errno));
+						log_sys_error("fclose", file);
 					return 1;
 				}
 				dm_bit_set(_dm_bitset, num);
@@ -160,7 +159,7 @@ static int _get_proc_number(const char *file, const char *name,
 		} while (c != EOF && c != '\n');
 	}
 	if (fclose(fl))
-		log_error("%s: fclose failed: %s", file, strerror(errno));
+		log_sys_error("fclose", file);
 
 	if (number) {
 		log_error("%s: No entry for %s found", file, name);
@@ -190,8 +189,7 @@ static int _control_exists(const char *control, uint32_t major, uint32_t minor)
 
 	if (stat(control, &buf) < 0) {
 		if (errno != ENOENT)
-			log_error("%s: stat failed: %s", control,
-				  strerror(errno));
+			log_sys_error("stat", control);
 		return 0;
 	}
 
@@ -199,8 +197,7 @@ static int _control_exists(const char *control, uint32_t major, uint32_t minor)
 		log_verbose("%s: Wrong inode type", control);
 		if (!unlink(control))
 			return 0;
-		log_error("%s: unlink failed: %s", control,
-			  strerror(errno));
+		log_sys_error("unlink", control);
 		return -1;
 	}
 
@@ -211,8 +208,7 @@ static int _control_exists(const char *control, uint32_t major, uint32_t minor)
 			    major, minor);
 		if (!unlink(control))
 			return 0;
-		log_error("%s: unlink failed: %s", control,
-			  strerror(errno));
+		log_sys_error("unlink", control);
 		return -1;
 	}
 
@@ -228,7 +224,7 @@ static int _create_control(const char *control, uint32_t major, uint32_t minor)
 		return 0;
 
 	old_umask = umask(0022);
-	ret = create_dir(dm_dir());
+	ret = dm_create_dir(dm_dir());
 	umask(old_umask);
 
 	if (!ret)
@@ -238,7 +234,7 @@ static int _create_control(const char *control, uint32_t major, uint32_t minor)
 
 	if (mknod(control, S_IFCHR | S_IRUSR | S_IWUSR,
 		  MKDEV(major, minor)) < 0)  {
-		log_error("%s: mknod failed: %s", control, strerror(errno));
+		log_sys_error("mknod", control);
 		return 0;
 	}
 
@@ -301,7 +297,7 @@ static int _open_control(void)
 		goto error;
 
 	if ((_control_fd = open(control, O_RDWR)) < 0) {
-		log_error("%s: open failed: %s", control, strerror(errno));
+		log_sys_error("open", control);
 		goto error;
 	}
 
@@ -592,12 +588,12 @@ static int _dm_names_v1(struct dm_ioctl_v1 *dmi)
 	struct stat buf;
 	char path[PATH_MAX];
 
-	log_print("Warning: Device list may be incomplete with interface "
+	log_warn("WARNING: Device list may be incomplete with interface "
 		  "version 1.");
-	log_print("Please upgrade your kernel device-mapper driver.");
+	log_warn("Please upgrade your kernel device-mapper driver.");
 
 	if (!(d = opendir(dev_dir))) {
-		log_error("%s: opendir failed: %s", dev_dir, strerror(errno));
+		log_sys_error("opendir", dev_dir);
 		return 0;
 	}
 
@@ -616,7 +612,7 @@ static int _dm_names_v1(struct dm_ioctl_v1 *dmi)
 						      (void *) old_names);
 		snprintf(path, sizeof(path), "%s/%s", dev_dir, name);
 		if (stat(path, &buf)) {
-			log_error("%s: stat failed: %s", path, strerror(errno));
+			log_sys_error("stat", path);
 			continue;
 		}
 		if (!S_ISBLK(buf.st_mode))
@@ -637,7 +633,7 @@ static int _dm_names_v1(struct dm_ioctl_v1 *dmi)
 	}
 
 	if (closedir(d))
-		log_error("%s: closedir failed: %s", dev_dir, strerror(errno));
+		log_sys_error("closedir", dev_dir);
 
 	return r;
 }
@@ -693,7 +689,7 @@ static int _dm_task_run_v1(struct dm_task *dmt)
 
 	if (dmi->flags & DM_BUFFER_FULL_FLAG)
 		/* FIXME Increase buffer size and retry operation (if query) */
-		log_error("Warning: libdevmapper buffer too small for data");
+		log_error("WARNING: libdevmapper buffer too small for data");
 
 	switch (dmt->type) {
 	case DM_DEVICE_CREATE:
@@ -924,6 +920,33 @@ int dm_task_get_info(struct dm_task *dmt, struct dm_info *info)
 	return 1;
 }
 
+uint32_t dm_task_get_read_ahead(const struct dm_task *dmt, uint32_t *read_ahead)
+{              
+	const char *dev_name;
+
+	*read_ahead = 0;
+
+#ifdef DM_COMPAT
+	/* Not supporting this */
+        if (_dm_version == 1)
+                return 1;
+#endif  
+
+        if (!dmt->dmi.v4 || !(dmt->dmi.v4->flags & DM_EXISTS_FLAG))
+		return 0;
+
+	if (*dmt->dmi.v4->name)
+		dev_name = dmt->dmi.v4->name;
+	else if (dmt->dev_name)
+		dev_name = dmt->dev_name;
+	else {
+		log_error("Get read ahead request failed: device name unrecorded.");
+		return 0;
+	}
+
+	return get_dev_node_read_ahead(dev_name, read_ahead);
+}
+
 const char *dm_task_get_name(const struct dm_task *dmt)
 {
 #ifdef DM_COMPAT
@@ -978,6 +1001,15 @@ int dm_task_set_ro(struct dm_task *dmt)
 	return 1;
 }
 
+int dm_task_set_read_ahead(struct dm_task *dmt, uint32_t read_ahead,
+			   uint32_t read_ahead_flags)
+{
+	dmt->read_ahead = read_ahead;
+	dmt->read_ahead_flags = read_ahead_flags;
+
+	return 1;
+}
+
 int dm_task_suppress_identical_reload(struct dm_task *dmt)
 {
 	dmt->suppress_identical_reload = 1;
@@ -1271,7 +1303,9 @@ static struct dm_ioctl *_flatten(struct dm_task *dmt, unsigned repeat_count)
 			    dmt->major, dmt->minor, dmi->name);
 	}
 
-	if (dmt->dev_name)
+	/* FIXME Until resume ioctl supplies name, use dev_name for readahead */
+	if (dmt->dev_name && (dmt->type != DM_DEVICE_RESUME || dmt->minor < 0 ||
+			      dmt->major < 0))
 		strncpy(dmi->name, dmt->dev_name, sizeof(dmi->name));
 
 	if (dmt->uuid)
@@ -1324,7 +1358,7 @@ static int _process_mapper_dir(struct dm_task *dmt)
 
 	dir = dm_dir();
 	if (!(d = opendir(dir))) {
-		log_error("opendir %s: %s", dir, strerror(errno));
+		log_sys_error("opendir", dir);
 		return 0;
 	}
 
@@ -1338,7 +1372,7 @@ static int _process_mapper_dir(struct dm_task *dmt)
 	}
 
 	if (closedir(d))
-		log_error("closedir %s: %s", dir, strerror(errno));
+		log_sys_error("closedir", dir);
 
 	return r;
 }
@@ -1650,14 +1684,16 @@ repeat_ioctl:
 			dm_free(dmi);
 			goto repeat_ioctl;
 		default:
-			log_error("Warning: libdevmapper buffer too small for data");
+			log_error("WARNING: libdevmapper buffer too small for data");
 		}
 	}
 
 	switch (dmt->type) {
 	case DM_DEVICE_CREATE:
-		add_dev_node(dmt->dev_name, MAJOR(dmi->dev), MINOR(dmi->dev),
-			     dmt->uid, dmt->gid, dmt->mode);
+		if (dmt->dev_name && *dmt->dev_name)
+			add_dev_node(dmt->dev_name, MAJOR(dmi->dev),
+				     MINOR(dmi->dev), dmt->uid, dmt->gid,
+				     dmt->mode);
 		break;
 
 	case DM_DEVICE_REMOVE:
@@ -1672,6 +1708,12 @@ repeat_ioctl:
 			rename_dev_node(dmt->dev_name, dmt->newname);
 		break;
 
+	case DM_DEVICE_RESUME:
+		/* FIXME Kernel needs to fill in dmi->name */
+		set_dev_node_read_ahead(dmt->dev_name, dmt->read_ahead,
+					dmt->read_ahead_flags);
+		break;
+	
 	case DM_DEVICE_MKNODES:
 		if (dmi->flags & DM_EXISTS_FLAG)
 			add_dev_node(dmi->name, MAJOR(dmi->dev),
